package meerkat.bulletinboard;

import com.google.common.util.concurrent.FutureCallback;
import com.google.common.util.concurrent.Futures;
import com.google.common.util.concurrent.ListeningScheduledExecutorService;
import com.google.common.util.concurrent.MoreExecutors;
import com.google.protobuf.Int64Value;
import com.google.protobuf.Timestamp;
import meerkat.bulletinboard.workers.singleserver.*;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Voting.BulletinBoardClientParams;
import meerkat.util.BulletinBoardUtils;

import java.lang.Iterable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Created by Arbel Deutsch Peled on 28-Dec-15.
 *
 * This class implements the asynchronous Bulletin Board Client interface
 * It only handles a single Bulletin Board Server
 * If the list of servers contains more than one server: the server actually used is the first one
 * The class further implements a delayed access to the server after a communication error occurs
 */
public class SingleServerBulletinBoardClient extends SimpleBulletinBoardClient implements SubscriptionBulletinBoardClient {

    private final int MAX_RETRIES = 11;

    private ListeningScheduledExecutorService executorService;

    private long lastServerErrorTime;

    private final long failDelayInMilliseconds;

    private final long subscriptionIntervalInMilliseconds;

    /**
     * Notify the client that a job has failed
     * This makes new scheduled jobs be scheduled for a later time (after the given delay)
     */
    protected void fail() {

        // Update last fail time
        lastServerErrorTime = System.currentTimeMillis();

    }

    /**
     * This method adds a worker to the scheduled queue of the threadpool
     * If the server is in an accessible state: the job is submitted for immediate handling
     * If the server is not accessible: the job is scheduled for a later time
     * @param worker is the worker that should be scheduled for work
     * @param callback is the class containing callbacks for handling job completion/failure
     */
    protected void scheduleWorker(SingleServerWorker worker, FutureCallback callback){

        long timeSinceLastServerError = System.currentTimeMillis() - lastServerErrorTime;

        if (timeSinceLastServerError >= failDelayInMilliseconds) {

            // Schedule for immediate processing
            Futures.addCallback(executorService.submit(worker), callback);

        } else {

            // Schedule for processing immediately following delay expiry
            Futures.addCallback(executorService.schedule(
                    worker,
                    failDelayInMilliseconds - timeSinceLastServerError,
                    TimeUnit.MILLISECONDS),
                    callback);

        }

    }

    /**
     * Inner class for handling simple operation results and retrying if needed
     */
    class RetryCallback<T> implements FutureCallback<T> {

        private final SingleServerWorker worker;
        private final FutureCallback<T> futureCallback;

        public RetryCallback(SingleServerWorker worker, FutureCallback<T> futureCallback) {
            this.worker = worker;
            this.futureCallback = futureCallback;
        }

        @Override
        public void onSuccess(T result) {
            if (futureCallback != null)
                futureCallback.onSuccess(result);
        }

        @Override
        public void onFailure(Throwable t) {

            // Notify client about failure
            fail();

            // Check if another attempt should be made

            worker.decMaxRetry();

            if (worker.isRetry()) {
                // Perform another attempt
                scheduleWorker(worker, this);
            } else {
                // No more retries: notify caller about failure
                if (futureCallback != null)
                    futureCallback.onFailure(t);
            }

        }

    }

    /**
     * This callback ties together all the per-batch-data callbacks into a single callback
     * It reports success back to the user only if all of the batch-data were successfully posted
     * If any batch-data fails to post: this callback reports failure
     */
    class PostBatchChunkListCallback implements FutureCallback<Boolean> {

        private final FutureCallback<Boolean> callback;

        private AtomicInteger batchDataRemaining;
        private AtomicBoolean aggregatedResult;

        public PostBatchChunkListCallback(int batchDataLength, FutureCallback<Boolean> callback) {

            this.callback = callback;
            this.batchDataRemaining = new AtomicInteger(batchDataLength);
            this.aggregatedResult = new AtomicBoolean(false);

        }

        @Override
        public void onSuccess(Boolean result) {

            if (result){
                this.aggregatedResult.set(true);
            }

            if (batchDataRemaining.decrementAndGet() == 0){
                if (callback != null)
                    callback.onSuccess(this.aggregatedResult.get());
            }
        }

        @Override
        public void onFailure(Throwable t) {

            // Notify caller about failure
            if (callback != null)
                callback.onFailure(t);

        }
    }

    /**
     * This callback ties together the different parts of a CompleteBatch as they arrive from the server
     * It assembles a CompleteBatch from the parts and sends it to the user if all parts arrived
     * If any part fails to arrive: it invokes the onFailure method
     */
    class CompleteBatchReadCallback {

        private final FutureCallback<BulletinBoardMessage> callback;

        private List<BatchChunk> batchChunkList;
        private BulletinBoardMessage stub;

        private AtomicInteger remainingQueries;
        private AtomicBoolean failed;

        public CompleteBatchReadCallback(FutureCallback<BulletinBoardMessage> callback) {

            this.callback = callback;

            remainingQueries = new AtomicInteger(2);
            failed = new AtomicBoolean(false);

        }

        protected void combineAndReturn() {

            if (remainingQueries.decrementAndGet() == 0){

                if (callback != null)
                    callback.onSuccess(BulletinBoardUtils.gatherBatch(stub, batchChunkList));
            }

        }

        protected void fail(Throwable t) {
            if (failed.compareAndSet(false, true)) {
                if (callback != null)
                    callback.onFailure(t);
            }
        }

        /**
         * @return a FutureCallback for the Batch Data List that ties to this object
         */
        public FutureCallback<List<BatchChunk>> asBatchDataListFutureCallback() {
            return new FutureCallback<List<BatchChunk>>() {

                @Override
                public void onSuccess(List<BatchChunk> result) {
                    batchChunkList = result;

                    combineAndReturn();
                }

                @Override
                public void onFailure(Throwable t) {
                    fail(t);
                }

            };
        }

        /**
         * @return a FutureCallback for the Bulletin Board Message that ties to this object
         */
        public FutureCallback<List<BulletinBoardMessage>> asBulletinBoardMessageListFutureCallback() {
            return new FutureCallback<List<BulletinBoardMessage>>() {

                @Override
                public void onSuccess(List<BulletinBoardMessage> result) {
                    if (result.size() < 1){
                        onFailure(new IllegalArgumentException("Server returned empty message list"));
                        return;
                    }

                    stub = result.get(0);

                    combineAndReturn();
                }

                @Override
                public void onFailure(Throwable t) {
                    fail(t);
                }
            };
        }

    }


    /**
     * Inner class for handling returned values of subscription operations
     * This class's methods also ensure continued operation of the subscription
     */
    class SubscriptionCallback implements FutureCallback<List<BulletinBoardMessage>> {

        private SingleServerReadMessagesWorker worker;
        private final FutureCallback<List<BulletinBoardMessage>> callback;

        private MessageFilterList.Builder filterBuilder;

        public SubscriptionCallback(SingleServerReadMessagesWorker worker, FutureCallback<List<BulletinBoardMessage>> callback) {
            this.worker = worker;
            this.callback = callback;
            filterBuilder = worker.getPayload().toBuilder();

        }

        @Override
        public void onSuccess(List<BulletinBoardMessage> result) {

            // Report new messages to user
            if (callback != null)
                callback.onSuccess(result);

            // Remove last filter from list (MIN_ENTRY one)
            filterBuilder.removeFilter(filterBuilder.getFilterCount() - 1);

            // Add updated MIN_ENTRY filter (entry number is successor of last received entry's number)
            filterBuilder.addFilter(MessageFilter.newBuilder()
                    .setType(FilterType.MIN_ENTRY)
                    .setEntry(result.get(result.size() - 1).getEntryNum() + 1)
                    .build());

            // Create new worker with updated task
            worker = new SingleServerReadMessagesWorker(worker.serverAddress, filterBuilder.build(), 1);

            // Schedule the worker
            scheduleWorker(worker, this);

        }

        @Override
        public void onFailure(Throwable t) {

            // Notify client about failure
            fail();

            // Notify caller about failure and terminate subscription
            if (callback != null)
                callback.onFailure(t);
        }
    }

    public SingleServerBulletinBoardClient(ListeningScheduledExecutorService executorService,
                                           long failDelayInMilliseconds,
                                           long subscriptionIntervalInMilliseconds) {

        this.executorService = executorService;

        this.failDelayInMilliseconds = failDelayInMilliseconds;
        this.subscriptionIntervalInMilliseconds = subscriptionIntervalInMilliseconds;

        // Set server error time to a time sufficiently in the past to make new jobs go through
        lastServerErrorTime = System.currentTimeMillis() - failDelayInMilliseconds;

    }

    public SingleServerBulletinBoardClient(int threadPoolSize, long failDelayInMilliseconds, long subscriptionIntervalInMilliseconds) {

        this(MoreExecutors.listeningDecorator(Executors.newScheduledThreadPool(threadPoolSize)),
                failDelayInMilliseconds,
                subscriptionIntervalInMilliseconds);

    }

    /**
     * Stores database location, initializes the web Client and
     * @param clientParams contains the data needed to access the DBs
     */
    @Override
    public void init(BulletinBoardClientParams clientParams) {

        // Perform usual setup
        super.init(clientParams);

        // Remove all but first DB address
        String dbAddress = meerkatDBs.get(0);
        meerkatDBs = new LinkedList<>();
        meerkatDBs.add(dbAddress);

    }

    @Override
    public MessageID postMessage(BulletinBoardMessage msg, FutureCallback<Boolean> callback) {

        // Create worker with redundancy 1 and MAX_RETRIES retries
        SingleServerPostMessageWorker worker = new SingleServerPostMessageWorker(meerkatDBs.get(0), msg, MAX_RETRIES);

        // Submit worker and create callback
        scheduleWorker(worker, new RetryCallback<>(worker, callback));

        // Calculate the correct message ID and return it
        digest.reset();
        digest.update(msg.getMsg());
        return digest.digestAsMessageID();

    }

    private class PostBatchDataCallback implements FutureCallback<Boolean> {

        private final BulletinBoardMessage msg;
        private final BatchIdentifier identifier;
        private final FutureCallback<Boolean> callback;

        public PostBatchDataCallback(BulletinBoardMessage msg, BatchIdentifier identifier, FutureCallback<Boolean> callback) {
            this.msg = msg;
            this.identifier = identifier;
            this.callback = callback;
        }

        @Override
        public void onSuccess(Boolean result) {
            closeBatch(
                    identifier,
                    msg.getMsg().getTimestamp(),
                    msg.getSigList(),
                    callback
            );
        }

        @Override
        public void onFailure(Throwable t) {
            if (callback != null)
                callback.onFailure(t);
        }

    }

    private class ContinueBatchCallback implements FutureCallback<BatchIdentifier> {

        private final BulletinBoardMessage msg;
        private final int chunkSize;
        private final FutureCallback<Boolean> callback;

        public ContinueBatchCallback(BulletinBoardMessage msg, int chunkSize, FutureCallback<Boolean> callback) {
            this.msg = msg;
            this.chunkSize = chunkSize;
            this.callback = callback;
        }

        @Override
        public void onSuccess(BatchIdentifier identifier) {

            List<BatchChunk> batchChunkList = BulletinBoardUtils.breakToBatch(msg, chunkSize);

            postBatchData(
                    identifier,
                    batchChunkList,
                    0,
                    new PostBatchDataCallback(msg, identifier, callback));
        }

        @Override
        public void onFailure(Throwable t) {
            if (callback != null)
                callback.onFailure(t);
        }
    }

    @Override
    public MessageID postAsBatch(BulletinBoardMessage msg, int chunkSize, FutureCallback<Boolean> callback) {

        beginBatch(
                msg.getMsg().getTagList(),
                new ContinueBatchCallback(msg, chunkSize, callback)
        );

        digest.update(msg);

        return digest.digestAsMessageID();

    }

    private class BeginBatchCallback implements FutureCallback<Int64Value> {

        private final FutureCallback<BatchIdentifier> callback;

        public BeginBatchCallback(FutureCallback<BatchIdentifier> callback) {
            this.callback = callback;
        }

        @Override
        public void onSuccess(Int64Value result) {
            callback.onSuccess(new SingleServerBatchIdentifier(result));
        }

        @Override
        public void onFailure(Throwable t) {
            callback.onFailure(t);
        }
    }

    @Override
    public void beginBatch(Iterable<String> tags, FutureCallback<BatchIdentifier> callback) {

        BeginBatchMessage beginBatchMessage = BeginBatchMessage.newBuilder()
                .addAllTag(tags)
                .build();

        // Create worker with redundancy 1 and MAX_RETRIES retries
        SingleServerBeginBatchWorker worker =
                new SingleServerBeginBatchWorker(meerkatDBs.get(0), beginBatchMessage, MAX_RETRIES);

        // Submit worker and create callback
        scheduleWorker(worker, new RetryCallback<>(worker, new BeginBatchCallback(callback)));

    }


    @Override
    public void postBatchData(BatchIdentifier batchIdentifier, List<BatchChunk> batchChunkList,
                          int startPosition, FutureCallback<Boolean> callback) throws IllegalArgumentException{

        // Cast identifier to usable form

        if (!(batchIdentifier instanceof SingleServerBatchIdentifier)){
            throw new IllegalArgumentException("Error: batch identifier supplied was not created by this class.");
        }

        SingleServerBatchIdentifier identifier = (SingleServerBatchIdentifier) batchIdentifier;

        // Update batch size

        identifier.setLength(startPosition + batchChunkList.size());

        // Create a unified callback to aggregate successful posts

        PostBatchChunkListCallback listCallback = new PostBatchChunkListCallback(batchChunkList.size(), callback);

        // Iterate through data list

        BatchMessage.Builder builder = BatchMessage.newBuilder()
                .setBatchId(identifier.getBatchId().getValue());

        for (BatchChunk data : batchChunkList) {
            builder.setSerialNum(startPosition).setData(data);

            // Create worker with redundancy 1 and MAX_RETRIES retries
            SingleServerPostBatchWorker worker =
                    new SingleServerPostBatchWorker(meerkatDBs.get(0), builder.build(), MAX_RETRIES);

            // Create worker with redundancy 1 and MAX_RETRIES retries
            scheduleWorker(worker, new RetryCallback<>(worker, listCallback));

            // Increment position in batch
            startPosition++;
        }

    }

    @Override
    public void postBatchData(BatchIdentifier batchIdentifier, List<BatchChunk> batchChunkList, FutureCallback<Boolean> callback)
        throws IllegalArgumentException {

        postBatchData(batchIdentifier, batchChunkList, 0, callback);

    }

    @Override
    public void closeBatch(BatchIdentifier batchIdentifier, Timestamp timestamp, Iterable<Crypto.Signature> signatures, FutureCallback<Boolean> callback)
            throws IllegalArgumentException {

        if (!(batchIdentifier instanceof SingleServerBatchIdentifier)){
            throw new IllegalArgumentException("Error: batch identifier supplied was not created by this class.");
        }

        SingleServerBatchIdentifier identifier = (SingleServerBatchIdentifier) batchIdentifier;

        CloseBatchMessage closeBatchMessage = CloseBatchMessage.newBuilder()
                .setBatchId(identifier.getBatchId().getValue())
                .setBatchLength(identifier.getLength())
                .setTimestamp(timestamp)
                .addAllSig(signatures)
                .build();

        // Create worker with redundancy 1 and MAX_RETRIES retries
        SingleServerCloseBatchWorker worker =
                new SingleServerCloseBatchWorker(meerkatDBs.get(0), closeBatchMessage, MAX_RETRIES);

        // Submit worker and create callback
        scheduleWorker(worker, new RetryCallback<>(worker, callback));

    }

    @Override
    public void getRedundancy(MessageID id, FutureCallback<Float> callback) {

        // Create worker with no retries
        SingleServerGetRedundancyWorker worker = new SingleServerGetRedundancyWorker(meerkatDBs.get(0), id, 1);

        // Submit job and create callback
        scheduleWorker(worker, new RetryCallback<>(worker, callback));

    }

    @Override
    public void readMessages(MessageFilterList filterList, FutureCallback<List<BulletinBoardMessage>> callback) {

        // Create job with no retries
        SingleServerReadMessagesWorker worker = new SingleServerReadMessagesWorker(meerkatDBs.get(0), filterList, 1);

        // Submit job and create callback
        scheduleWorker(worker, new RetryCallback<>(worker, callback));

    }

    @Override
    public void readBatch(MessageID msgID, FutureCallback<BulletinBoardMessage> callback) {

        // Create job with MAX retries for retrieval of the Bulletin Board Message that defines the batch

        MessageFilterList filterList = MessageFilterList.newBuilder()
                .addFilter(MessageFilter.newBuilder()
                        .setType(FilterType.MSG_ID)
                        .setId(msgID.getID())
                        .build())
                .build();

        BatchQuery batchQuery = BatchQuery.newBuilder()
                .setMsgID(msgID)
                .setStartPosition(0)
                .build();

        SingleServerReadMessagesWorker messageWorker = new SingleServerReadMessagesWorker(meerkatDBs.get(0), filterList, MAX_RETRIES);

        // Create job with MAX retries for retrieval of the Batch Data List
        SingleServerReadBatchWorker batchWorker = new SingleServerReadBatchWorker(meerkatDBs.get(0), batchQuery, MAX_RETRIES);

        // Create callback that will combine the two worker products
        CompleteBatchReadCallback completeBatchReadCallback = new CompleteBatchReadCallback(callback);

        // Submit jobs with wrapped callbacks
        scheduleWorker(messageWorker, new RetryCallback<>(messageWorker, completeBatchReadCallback.asBulletinBoardMessageListFutureCallback()));
        scheduleWorker(batchWorker, new RetryCallback<>(batchWorker, completeBatchReadCallback.asBatchDataListFutureCallback()));

    }

    private class ReadBatchCallback implements FutureCallback<List<BatchChunk>> {

        private final BulletinBoardMessage stub;
        private final FutureCallback<BulletinBoardMessage> callback;

        public ReadBatchCallback(BulletinBoardMessage stub, FutureCallback<BulletinBoardMessage> callback) {
            this.stub = stub;
            this.callback = callback;
        }

        @Override
        public void onSuccess(List<BatchChunk> result) {
            callback.onSuccess(BulletinBoardUtils.gatherBatch(stub, result));
        }

        @Override
        public void onFailure(Throwable t) {

        }
    }

    @Override
    public void readBatchData(BulletinBoardMessage stub, FutureCallback<BulletinBoardMessage> callback) throws IllegalArgumentException{

        if (stub.getMsg().getDataTypeCase() != UnsignedBulletinBoardMessage.DataTypeCase.MSGID) {
            throw new IllegalArgumentException("Message is not a stub and does not contain the required message ID");
        }

        // Create job with MAX retries for retrieval of the Batch Data List

        BatchQuery batchQuery = BatchQuery.newBuilder()
                .setMsgID(MessageID.newBuilder()
                        .setID(stub.getMsg().getMsgId())
                        .build())
                .setStartPosition(0)
                .build();

        SingleServerReadBatchWorker batchWorker = new SingleServerReadBatchWorker(meerkatDBs.get(0), batchQuery, MAX_RETRIES);

        scheduleWorker(batchWorker, new RetryCallback<>(batchWorker, new ReadBatchCallback(stub, callback)));

    }

    @Override
    public void querySync(SyncQuery syncQuery, FutureCallback<SyncQueryResponse> callback) {

        SingleServerQuerySyncWorker worker = new SingleServerQuerySyncWorker(meerkatDBs.get(0), syncQuery, MAX_RETRIES);

        scheduleWorker(worker, new RetryCallback<>(worker, callback));

    }

    @Override
    public void subscribe(MessageFilterList filterList, long startEntry, FutureCallback<List<BulletinBoardMessage>> callback) {
        // Remove all existing MIN_ENTRY filters and create new one that starts at 0

        MessageFilterList.Builder filterListBuilder = filterList.toBuilder();

        Iterator<MessageFilter> iterator = filterListBuilder.getFilterList().iterator();
        while (iterator.hasNext()) {
            MessageFilter filter = iterator.next();

            if (filter.getType() == FilterType.MIN_ENTRY){
                iterator.remove();
            }
        }
        filterListBuilder.addFilter(MessageFilter.newBuilder()
                .setType(FilterType.MIN_ENTRY)
                .setEntry(startEntry)
                .build());

        // Create job with no retries
        SingleServerReadMessagesWorker worker = new SingleServerReadMessagesWorker(meerkatDBs.get(0), filterListBuilder.build(), MAX_RETRIES);

        // Submit job and create callback that retries on failure and handles repeated subscription
        scheduleWorker(worker, new RetryCallback<>(worker, new SubscriptionCallback(worker, callback)));
    }

    @Override
    public void subscribe(MessageFilterList filterList, FutureCallback<List<BulletinBoardMessage>> callback) {
        subscribe(filterList, 0, callback);
    }

    @Override
    public void close() {

        super.close();

        executorService.shutdown();

    }
}
